/*		 ITI-FG  Kernel module free public release
 *
 *		     Copyright (C) 1996-2000 GOM mbH.
 *	         Rebenring 33, 38106 Braunschweig, GERMANY
 *		            All rights reserved
 *
 * $Id: convert.c,v 7.0 2000/10/25 09:07:57 cvs Exp $
 *
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>

#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/ioctl.h>

#ifdef linux
#  include <linux/limits.h>	/* PATH_MAX */
#endif

#include "itifgExt.h"
#include "iticonfig.h"

#ifndef MAP_FAILED
#  define MAP_FAILED	((void*)-1)
#endif

int
convert_fpgafile (const char* config_file, const char* device_file)
{
  int config_desc = -1, device_desc = -1;
  struct stat config_stat;
  size_t config_size = 0;

  caddr_t data_in = NULL;
  caddr_t token;
  caddr_t endptr;
  char convert[7];

  size_t length;

  int addrlen;
  int datalen;
  int offset;

  u_char value;
  u_char data_out[PCD_CONFIG_BYTES];

  int error;
  int bytes;
  int i;


  if (config_file == NULL || device_file == NULL)
    {
      error = -EITI_ARG;
      goto cleanup;
    }
  if ((config_desc = open (config_file, O_RDONLY)) < 0)
    {
      if (errno == ENOENT)
	{
	  error = -EITI_ENT;
	  goto cleanup;
	}
      else
	{
	  error = -EITI_SYS;
	  goto cleanup;
	}
    }

  if ((device_desc = open (device_file, O_WRONLY)) < 0)
    {
      if (errno == ENXIO)
	{
	  error = -EITI_ARG;
	  goto cleanup;
	}
      else
	{
	  error = -EITI_SYS;
	  goto cleanup;
	}
    }

  if (fstat (config_desc, &config_stat) < 0)
    {
      error = -EITI_SYS;
      goto cleanup;
    }
  config_size = config_stat.st_size;
  if ((data_in = mmap (NULL, config_size, PROT_READ | PROT_WRITE,
		       MAP_PRIVATE, config_desc, 0)) == MAP_FAILED)
    {
      error = -EITI_SYS;
      goto cleanup;
    }

  if ((token = strtok (data_in, "\n")) == NULL)
    {
      ITI_PRINTF ("Data currupted (no valid line in file).\n");
      error = -EITI_FMT;
      goto cleanup;
    }
  bytes = 0;
  while (TRUE)
    {
      length = strlen (token);
      if (length < 4)
	{
	  ITI_PRINTF ("Data currupted (no line header).\n");
	  error = -EITI_FMT;
	  goto cleanup;
	}
      if (token[0] != 'S')
	{
	  ITI_PRINTF ("Data currupted (no magic char found).\n");
	  error = -EITI_FMT;
	  goto cleanup;
	}
      switch (addrlen = token[1])
	{
	case '1':
	  addrlen = 2;
	  break;
	case '2':
	  addrlen = 3;
	  break;
	default:
	  ITI_PRINTF ("Data currupted (address length invalid).\n");
	  error = -EITI_FMT;
	  goto cleanup;
	}

      strncpy (convert, &token[2], 2);
      convert[2] = '\0';
      datalen = strtol (convert, &endptr, 16);
      if (*convert == '\0' || *endptr != '\0')
	{
	  ITI_PRINTF ("Data currupted (line length invalid).\n");
	  error = -EITI_FMT;
	  goto cleanup;
	}
      if ((addrlen == 2 && datalen < 2) || (addrlen == 3 && datalen < 3))
	{
	  ITI_PRINTF ("Data currupted (line length to short).\n");
	  error = -EITI_FMT;
	  goto cleanup;
	}
      ITI_PRINTF("datalen: %d", datalen);

      strncpy (convert, &token[4], addrlen * 2);
      convert[addrlen * 2] = '\0';
      offset = strtol (convert, &endptr, 16);
      if (*convert == '\0' || *endptr != '\0')
	{
	  ITI_PRINTF ("Data currupted (line number invalid).\n");
	  error = -EITI_FMT;
	  goto cleanup;
	}
      if (offset < 0)
	{
	  ITI_PRINTF ("Data currupted (line number to small).\n");
	  error = -EITI_FMT;
	  goto cleanup;
	}
      ITI_PRINTF(", offset: %d |", offset);

      if (addrlen == 2)
	i = 4;
      else
	i = 5;
      datalen--;	/* don't read the checksum */
      while (i < datalen + 2) /* but start at index 4 */
	{
	  strncpy (convert, &token[i * 2], 2);
	  convert[2] = '\0';
	  value = strtol (convert, &endptr, 16);
	  if (*convert == '\0' || *endptr != '\0')
	    {
	      ITI_PRINTF ("Data currupted (byte value invalid).\n");
	      error = -EITI_FMT;
	      goto cleanup;
	    }
	  ITI_PRINTF(" %x", value);
	  if (bytes >= PCD_CONFIG_BYTES)
	    break;
	  data_out[bytes] = value;
	  i++, bytes++;
	}
      ITI_PRINTF (" | (%d).\n", bytes);
      
      if (bytes >= PCD_CONFIG_BYTES)
	break;
      if ((token = strtok (NULL, "\n")) == NULL)
	{
	  ITI_PRINTF ("Data currupted (no more lines in file).\n");
	  error = -EITI_FMT;
	  goto cleanup;
	}
    }

  if ((token = strtok (NULL, "\n")) == NULL)
    {
      ITI_PRINTF ("Data currupted (no more lines in file).\n");
      error = -EITI_FMT;
      goto cleanup;
    }
  length = strlen (token);
  if (length < 4)
    {
      ITI_PRINTF ("Data currupted (no line header).\n");
      error = -EITI_FMT;
      goto cleanup;
    }
  if (token[0] != 'S')
    {
      ITI_PRINTF ("Data currupted (no magic char found).\n");
      error = -EITI_FMT;
      goto cleanup;
    }
  if (token[1] != '9')
    {
      ITI_PRINTF ("Data currupted (extra data detected).\n");
      error = -EITI_FMT;
      goto cleanup;
    }

  ITI_PRINTF ("%s: %d bytes read.\n", config_file, bytes);
  
  if ((bytes = write (device_desc, data_out, PCD_CONFIG_BYTES))
      != PCD_CONFIG_BYTES)
    {
      ITI_PRINTF ("%s: %d bytes write.\n", device_file, bytes);
      error = -EITI_SYS;
      goto cleanup;
    }
  
  ITI_PRINTF ("%s: %d bytes write.\n", device_file, bytes);

  if (munmap (data_in, config_size) < 0)
    return -EITI_SYS;
  if (close (device_desc) < 0)
    return -EITI_SYS;
  if (close (config_desc) < 0)
    return -EITI_SYS;

  return OK;

 cleanup:
  if (data_in != NULL && config_size != 0)
    munmap (data_in, config_size);
  if (device_desc != -1)
    close (device_desc);
  if (config_desc != -1)
    close (config_desc);
  return error;
}
